# src/mulch/workspace_manager_generator.py (project-agnostic, reusable)

import json
import logging
from pathlib import Path
from jinja2 import Environment, PackageLoader, select_autoescape #,FileSystemLoader
import toml
import typer
from importlib.resources import files

from mulch.helpers import get_global_config_path, get_user_root, try_load_scaffold_file
from mulch.constants import FALLBACK_SCAFFOLD, DEFAULT_SCAFFOLD_FILENAME
from mulch.logging_setup import setup_logging, setup_logging_portable
from mulch.flags_log_status import FlagsLockStatus
from mulch.basepath_manager import PathContext


setup_logging_portable()
logger = logging.getLogger(__name__)


class WorkspaceManagerGenerator:
    f"""
    workspace manager generator for use with the mulch CLI command `mulch src`.
    Manages directory creation and standardized file placement based on {DEFAULT_SCAFFOLD_FILENAME}.
    """
    
    DEFAULT_WORKSPACE_CONFIG_FILENAME = "default-workspace.toml"
    DEFAULT_TEMPLATE_DIR = Path(__file__).parent / "templates"
    DEFAULT_WORKSPACE_TEMPLATE_FILENAME = "workspace_manager.py.j2"
    FALLBACK_SCAFFOLD = FALLBACK_SCAFFOLD # to make accessible, for pip and interally
    DEFAULT_SCAFFOLD_FILENAME = DEFAULT_SCAFFOLD_FILENAME # to make accessible, for pip and interally
    
    def __init__(self, base_path: Path, project_name: str, lock_data: dict, stealth: bool = False, force: bool = False):
        """
        WorkspaceManagerGenerator.context is a mystery to me.
        """
        self.base_path = Path(base_path).resolve()
        self.lock_data = lock_data
        self.stealth = stealth 
        self.force = force
        self.context = PathContext(base_path = base_path, 
                                   project_name = project_name, 
                                   workspace_name=None, 
                                   here=None, 
                                   stealth=stealth)
        self.flags_lock_path = self.context.flags_lock_path
        self.manager_lock_path = self.context.manager_lock_path
        self.src_path = self.manager_lock_path.parent
        self.manager_path = self.context.manager_path 
        #self.project_name = self.base_path.name # assumption that the target dir is the package name, fair enough
        self.project_name = project_name

    def get_path(self, key: str) -> Path:
        """
        Generic path getter using slash-separated key within the workspace.
        """
        path = self.workspace_dir
        for part in key.strip("/").split("/"):
            path /= part
        return path
    def evaluate_flags_lock_status(self) -> FlagsLockStatus:
        if not self.flags_lock_path.exists():
            return FlagsLockStatus.MISSING
        else:
            try:
                with open(self.flags_lock_path, "r", encoding="utf-8") as f:
                    existing = toml.loads(f)
            except:
                pass
        
    def build_src_components(self):
        setup_logging()
        was_source_generated = self.render_workspace_manager()
        return was_source_generated

    def render_workspace_manager(self):
        """
        Render a workspace_manager.py file based on the scaffold and template.
        """
        # jinja2 template loader from the mulch sourcecode
        env = Environment(
            loader=PackageLoader("mulch", "templates"),
            autoescape=select_autoescape()
        )
        template = env.get_template(self.DEFAULT_WORKSPACE_TEMPLATE_FILENAME)

        rendered = template.render(
            project_name = self.project_name,
            scaffold=self.lock_data["scaffold"]
        ) # self.lock_data["scaffold"] is the scaffold dict loaded from the mulch-scaffold file

        logger.info(f"src lock_path = {self.manager_lock_path}")
        if self.manager_lock_path.exists() and not self.force:
            try:
                with open(self.manager_lock_path, "r", encoding="utf-8") as f:
                    existing = json.load(f)
                existing_scaffold = existing.get("scaffold", {})
                if existing_scaffold == self.lock_data["scaffold"]: #self.scaffold:
                    logging.debug(f"Scaffold unchanged. Skipping re-render of workspace_manager.py at {self.manager_path}")
                    typer.echo(f"Scaffold unchanged. Skipping re-render of workspace_manager.py.")
                    return False # 🛑 Skip rendering
                else:
                    if not typer.confirm(f"⚠️ Existing {self.manager_lock_path} does not match this scaffold structure."
                                         "Overwriting the workspace_manager.py file can break references for existing workspaces. Continue?",
                                         abort=True):
                        logger.info("User chose not to overwrite existing workspace_manager.py")
                        return False
                    else: 
                        pass # 🟢 Continue rendering
            except Exception as e:
                logger.warning(f"Could not read {self.manager_lock_path.name} for comparison: {e}")
                return False
        # Only reach here if:
        # 1. No lock file exists
        # 2. User confirmed overwrite
        # 3. Scaffolds match
        self.manager_path.parent.mkdir(parents=True, exist_ok=True)
        self.manager_path.write_text(rendered)
        with open(self.manager_lock_path, "w", encoding="utf-8") as f:
            json.dump(self.lock_data, f, indent=2)
        typer.echo(f"workspace_manager.py generated!")
        logging.debug(f"Generated workspace_manager.py at {self.manager_path}")
        return True # 🟢 Indicate that rendering was successful


def load_scaffold(target_dir: Path | None = None, strict_local_dotmulch:bool=False, seed_if_missing:bool=False) -> dict:
    target_dir = target_dir or Path.cwd()
    base = target_dir / ".mulch"

    if strict_local_dotmulch:
        # Only try .mulch — no fallback
        for fname in filenames:
            path = base / fname
            scaffold = try_load_scaffold_file(path)
            if scaffold:
                logger.info(f"✅ Loaded scaffold from: {path}")
                return scaffold

        if seed_if_missing:
            from mulch.seed_logic import write_seed_scaffold  # or wherever this lives
            logger.warning("⚠️ .mulch exists but no scaffold file found. Auto-seeding...")
            write_seed_scaffold(target_dir)
            return load_scaffold(target_dir, strict_local_dotmulch=True, seed_if_missing=False)

        raise FileNotFoundError("🚫 No valid `.mulch/mulch-scaffold.*` found and auto-seed not enabled.")

    # Default behavior: search all fallback paths
    
    base_dirs = [
        target_dir / ".mulch",    # 1. Local .mulch folder
        target_dir,               # 2. Root project dir
        Path.home() / 'mulch',               # 3. User root on system
        get_global_config_path(appname = "mulch") # 4. Global config
    ]
    
    filenames = ["mulch.toml"]
    
    for base in base_dirs:
        for filename in filenames:
            path = base / filename
            scaffold = try_load_scaffold_file(path)
            if scaffold:
                logger.info(f"✅ Loaded scaffold from: {path}")
                return scaffold
            
    logger.warning("No valid scaffold file found. Falling back to internal scaffold.")
    return FALLBACK_SCAFFOLD




# src/mulch/cli.py

import typer
import toml
from pathlib import Path
import logging
from enum import Enum
import datetime
from importlib.metadata import version, PackageNotFoundError
from pprint import pprint
from rich.table import Table
from rich.console import Console
import sys

from mulch.decorators import with_logging
from mulch.workspace_manager_generator import WorkspaceManagerGenerator
from mulch.workspace_instance_factory import WorkspaceInstanceFactory
from mulch.logging_setup import setup_logging, setup_logging_portable
from mulch.helpers import dedupe_paths, open_editor, calculate_nowtime_foldername, get_local_appdata_path, get_default_untitled_workspace_name_based_on_operating_system, get_global_config_path, index_to_letters, get_username_from_home_directory
#from mulch.commands.dotfolder import create_dot_mulch
#from mulch.commands.build_dotmulch_standard_contents import build_dotmulch_standard_contents
from mulch.constants import FALLBACK_SCAFFOLD, LOCK_FILE_NAME, DEFAULT_SCAFFOLD_FILENAME
from mulch.workspace_status import WorkspaceStatus
from mulch.scaffold_loader import load_scaffold_file, resolve_scaffold
from mulch.reference_lock_manager import ReferenceLockManager, build_flags_record


logging.basicConfig(level=logging.INFO, format="%(levelname)s: %(message)s")
logger = logging.getLogger(__name__)


HELP_TEXT = "Mulch CLI for scaffolding Python project workspaces."
SCAFFOLD_TEMPLATES_FILENAME = 'mulch-scaffold-template-dictionary.toml'

FILENAMES_OF_RESPECT = [
    'mulch.toml'
]

# Paths are checked in order of respect for loading the scaffold template dictionary.
# These are used when running `mulch file`, `mulch show`, or any command needing layout templates.
ORDER_OF_RESPECT = [
    Path('.mulch'), # project local hidden folder
    Path.home() / '.mulch', # user profile hidden folde
    get_local_appdata_path("mulch"), # OS standard config path
    get_global_config_path("mulch") # Roaming AppData on Windows, ~/.config on Linux/macOS
]
maybe_global = get_global_config_path(appname="mulch")
if maybe_global:
    ORDER_OF_RESPECT.append(Path(maybe_global))

TEMPLATE_CHOICE_DICTIONARY_FILEPATHS = [
    p / SCAFFOLD_TEMPLATES_FILENAME
    for p in ORDER_OF_RESPECT
    if isinstance(p, Path)
]


try:
    MULCH_VERSION = version("mulch")
    __version__ = version("mulch")
except PackageNotFoundError:
    MULCH_VERSION = "unknown"

try:
    from importlib.metadata import version
    __version__ = version("mulch")
except PackageNotFoundError:
    # fallback if running from source
    try:
        with open(Path(__file__).parent / "VERSION") as f:
            __version__ = f.read().strip()
    except FileNotFoundError:
        __version__ = "dev"
    
# load the fallback_scaffold to this file

# Create the Typer CLI app
app = typer.Typer(help=HELP_TEXT, no_args_is_help=True, add_completion=False)

#@app.callback(invoke_without_command=True)
@app.callback()
def main(
    version: bool = typer.Option(None, "--version", callback=lambda v: print_version(v), is_eager=True, help="Show the version and exit.")
):
    """
    Mulch CLI for scaffolding Python project workspaces
    """
    
    lock = ReferenceLockManager.load_lock()
    if lock is None:
        print("Warning: No reference.lock found — run src and workspace commands first.")
    else:
        pass
        # validate, warn, etc

def print_version(value: bool):
    if value:
        try:
            typer.secho(f"mulch {MULCH_VERSION}",fg=typer.colors.GREEN, bold=True)
        except PackageNotFoundError:
            typer.echo("Version info not found")
        raise typer.Exit()

def _all_order_of_respect_failed(order_of_respect):
    failed = True
    for path in order_of_respect:
        if Path(path).exists():
            failed = False
    return failed

def make_dot_mulch_folder(target_dir):
    return create_dot_mulch(target_dir, order_of_respect=ORDER_OF_RESPECT)

@app.command()
@with_logging
def src(
    target_dir: Path = typer.Option(Path.cwd(), "--target-dir", "-r", help="Target project root (defaults to current directory)."),
    #enforce_mulch_folder: bool = typer.Option(False,"--enforce-mulch-folder-only-no-fallback", "-e", help = "This is leveraged in the CLI call by the context menu Mulch command PS1 to ultimately mean 'If you run Mulch and there is no .mulch folder, one will be generated. If there is one, it will use the default therein.' "),
    stealth: bool = typer.Option(False, "--stealth", "-s", help="Put source files in .mulch/src/ instead of src/."),
    force: bool = typer.Option(False, "--force", help="Override existing, forced."),
    expliref: bool = typer.Option(None, "--wrkspc-in-root/--wrkspc-in-wrkspc", "-wr/-ww", help="Allows you to run src command without first running the workspace command and without a prompt for more input. ww correlates here flag, wr correlates with late thereof."),
    name: str = typer.Option(None, "--name", "-n", help="Name of the src dir to create, in src/.")
    ):
    """
    Build the workspace_manager.py file in the source code, using the mulch.toml structure or the fallback structure embedded in WorkspaceManagerGenerator.
    Establish a logs folder at root, with the logging.json file.
    """
    command_line = ""#.join(sys.argv) # broken currently
    command_line = "null"
    #typer.echo(f"command_line = {command_line}")
    flags = build_flags_record(expliref=expliref,force=force,stealth=stealth,name=name)
    if name is None:
        project_name = target_dir.name
    else: 
        project_name = name
    
    '''
    ## A Rule: Run `mulch workspace` at least once before running `mulch src`
    (particularly if you want to run `mulch workspace --here`)
    - For `mulch src` to make the proper references, if you plan to use `--here` flag, you must run `mulch workspace --here` at least once first, to contribute some truth to the `.mulch/reference.lock` file.
    - If you run `mulch src` without first running `mulch workspace`, the workspace_manager.py references will include to the standard /workspaces/ reference, as if the `--here` flag was not used.
    '''
    # Set and infer the embedded reference location for workspace_manager, in case you want to successfully run `mulch src` before 
    if expliref is not None: # Note that this is the src command, not the workspace command.
        ReferenceLockManager.update_lock_workspace(pathstr="null", command_line = command_line, flags=flags) # "null" as a string is acceptable for pathstr, because if no workspace registered yet, expliref provides a basis for the workspace_manager.py references to be generated.

    order_of_respect_local = ORDER_OF_RESPECT
    if _all_order_of_respect_failed(order_of_respect_local):
       make_dot_mulch_folder(target_dir = Path.cwd()) # uses the same logic as the `mulch workspace` command. The `mulch file` command must be run manually, for that behavior to be achieved but otherwise the default is the `.mulch` manifestation. This should contain a query tool to build a `mulch.toml` file is the user is not comfortable doingediting it themselves in a text editor.

    scaffold_data = resolve_scaffold(order_of_respect_local, FILENAMES_OF_RESPECT)
    
    # Create lock data
    lock_data = {   
        "scaffold": scaffold_data,
        "generated_by": f"mulch {MULCH_VERSION}",
        "generated_at": datetime.datetime.now(datetime.timezone.utc).isoformat(),
        "generated_by": get_username_from_home_directory()
    }
    
    #manager_status = wf.evaluate_manager_status() # check the lock file in src/-packagename-/mulch.lock, which correlates with the workspacemanager
    
    #--- TO DO
    # - generate default name to this level, based on the project folder (CWD) name, and allow for override if --name flag is provided
    # some of the language here is confusing, because the /src/ folder will be built in the cwd is the stealth  
    mgf = WorkspaceManagerGenerator(base_path = target_dir, project_name = project_name, lock_data = lock_data, stealth=stealth, force=force)
    was_source_generated = mgf.build_src_components()
    if was_source_generated:
        typer.secho(f"📁 Source code created", fg=typer.colors.BRIGHT_GREEN)    
        # create reference lock file contents for src
        flags = build_flags_record(expliref=expliref,force=force,stealth=stealth)
        ReferenceLockManager.update_lock_src(pathstr=str(mgf.src_path), command_line = command_line ,flags=flags) # convert Path to str for serialization, at the point of input for clarity. "null" is acceptable as a string.

class NamingPattern(str,Enum):
    date = "date"
    new = "new"

def get_folder_name(pattern: NamingPattern = 'date', base_name: str = "New workspace", workspaces_dir: Path = Path.cwd() / "workspaces") -> str:
    '''
    Dynamically generate a workspace folder name based on the chosen pattern.
    Implementation, if the '--name' flag is not used with `mulch workspace`:
     - Default to {date}, and then {date}b, {date}c, {date}d
     - If the '--pattern new' is used when calling `mulch workspace`, the generated name will be 'New workspace', then 'New workspace (2)', etc, if Windows.   
     - 'mulch workspace --pattern new --here' will be used as the default register context menu command for 'mulch workspace', using the mulch-workspace.reg file. 
    '''
    
    if pattern == NamingPattern.date:
        suffix_index = 0
        while True:
            if suffix_index == 0:
                folder_name = calculate_nowtime_foldername()
            else:
                # Skip 'a', start from 'b'
                suffix = index_to_letters(suffix_index + 1)
                folder_name = f"{calculate_nowtime_foldername()}{suffix}"
            if not (workspaces_dir / folder_name).exists(): # 
                return folder_name
            suffix_index += 1
    elif pattern == NamingPattern.new:
        # check for existing workspace folders to append (n) if necessary, like "New workspace (2)", to mimmick windows "New folder (2)" behavior.
        return get_default_untitled_workspace_name_based_on_operating_system(workspaces_dir)
        

@app.command()
@with_logging
def workspace(
    target_dir: Path = typer.Option(Path.cwd(), "--target-dir", "-r", help="Target project root (defaults to current directory)."),
    pattern: NamingPattern = typer.Option(None, "--pattern", "-p",  help = "Choose naming pattern: 'date' for YYY_MMMMM_DD, or 'new' for os-specific pattern like 'New workspace (n)'"),
    name: str = typer.Option(None, "--name", "-n", help="Name of the workspace to create."),
    here: bool = typer.Option(False, "--here", "-h", help="The new named workspace directory should be placed immediately in the current working directory, rather than nested within a `/workspaces/` directory."),
    set_default: bool = typer.Option(True, "--set-default/--no-set-default", help="Write default-workspace.toml"),
    #enforce_mulch_folder: bool = typer.Option(False,"--enforce-mulch-folder-only-no-fallback", "-e", help = "This is leveraged in the CLI call by the context menu Mulch command PS1 to ultimately mean 'If you run Mulch and there is no .mulch folder, one will be generated. If there is one, it will use the default therein.' "),
    ):
    """
    Initialize a new workspace folder, using the mulch.toml structure or the fallback structure embedded in WorkspaceManagerGenerator.
    """
    command_line = ""#.join(sys.argv) # broken currently
    command_line = "null"
    #typer.echo(f"command_line = {command_line}")
    # Provide instant feedback on the --here setting, if used.
    if here:
        logger.debug(f"`here`: True")
    
    # First determine workspaces directory
    workspaces_dir = WorkspaceInstanceFactory.determine_workspaces_dir(
        target_dir=target_dir,
        here=here
    )
    # Second determine the flags used from the command line and record them, before they are changed in steps Three and Four immediately below.
    flags = build_flags_record(here=here,name=name,pattern=pattern)
    # Third, use the default pattern, if the pattern flag was not used.
    if pattern is None:
        pattern = NamingPattern.date # only necessary if name is None
    # Fourth, assign the patterned name, if the name flag was not used to supply the name explicitly.
    if name is None:
        name=get_folder_name(pattern = pattern, workspaces_dir=workspaces_dir)
    
    order_of_respect_local = ORDER_OF_RESPECT
    if _all_order_of_respect_failed(order_of_respect_local):
       make_dot_mulch_folder(target_dir = Path.cwd()) # uses the same logic as the `mulch workspace` command. The `mulch file` command must be run manually, for that behavior to be achieved but otherwise the default is the `.mulch` manifestation. This should contain a query tool to build a `mulch.toml` file is the user is not comfortable doingediting it themselves in a text editor.

    scaffold_data = resolve_scaffold(order_of_respect_local, FILENAMES_OF_RESPECT)
    
    # Create lock data
    lock_data = {
        "scaffold": scaffold_data,
        "generated_by": f"mulch {MULCH_VERSION}",
        "generated_at": datetime.datetime.now(datetime.timezone.utc).isoformat(),
        "generated_by": get_username_from_home_directory()
    }
    
    logger.debug(f"workspace_dirs = {workspaces_dir}")
    wif = WorkspaceInstanceFactory(target_dir, workspaces_dir, name, lock_data, here=here)
    
    workspace_status = wif.evaluate_workspace_status()
    
    if workspace_status == WorkspaceStatus.MATCHES:
        typer.secho(f"✅ Workspace '{name}' is already set up in {workspaces_dir}", fg=typer.colors.GREEN)
        typer.echo("   (Scaffold unchanged. Nothing regenerated.)")
        raise typer.Exit()

    elif workspace_status == WorkspaceStatus.DIFFERS:
        typer.secho(f"⚠️  Workspace '{name}' already exists and scaffold has changed.", fg=typer.colors.YELLOW)
        if not typer.confirm("Overwrite existing workspace?", default=False):
            typer.secho("❌ Aborting.", fg=typer.colors.RED)
            raise typer.Exit()

    elif workspace_status == WorkspaceStatus.EXISTS_NO_LOCK:
        typer.secho(f"⚠️  Workspace exists at {workspaces_dir / name} but no scaffold.lock found.", fg=typer.colors.YELLOW)
        if not typer.confirm("Overwrite existing workspace?", default=False):
            typer.secho("❌ Aborting.", fg=typer.colors.RED)
            raise typer.Exit()
        
    # Proceed to generate
    wif.create_workspace(set_default=set_default)

    # create reference lock file contents for workspace 
    ReferenceLockManager.update_lock_workspace(pathstr=str(workspaces_dir / name), command_line = command_line, flags=flags)
    
@app.command()
def context():
    """
    Install the right-click `mulch workspace` context menu registry item.
    """
    from mulch.scripts.install import install_context
    install_context.setup()
    
def load_template_choice_dictionary_from_file():
    """
    Attempts to load a TOML or JSON template choice dictionary from known fallback paths.
    """
    for path in TEMPLATE_CHOICE_DICTIONARY_FILEPATHS:
        if path.is_file():
            data = load_scaffold_file(path)
            if data is not None:
                typer.secho(f"✅ Loaded template choices from: {path}", fg=typer.colors.GREEN)
                return data
            else:
                typer.secho(f"⚠️ Failed to parse {path.name} as TOML or JSON.", fg=typer.colors.YELLOW)
    typer.secho("❌ Failed to load template choice dictionary from any known paths.", fg=typer.colors.RED)
    raise typer.Exit(code=1)

@app.command()
def seed(
    target_dir: Path = typer.Option(Path.cwd(), "--target-dir", "-t", help="Target project root (defaults to current directory)."),
    index: int = typer.Option(None, "--index", "-i", help="Index from 'mulch order' to choose scaffold source."),
    template_choice: int = typer.Option(None, "--template-choice", "-c", help="Reference a known template for workspace organization."),
    edit: bool = typer.Option(False, "--edit", "-e", help="Open scaffold file for editing after creation."),
):
    """
    Drop a .mulch folder to disk at the target directory.
    The scaffold source can be selected by index (from 'mulch order') or by template choice.
    You can edit the .mulch/mulch.toml file manually. Coming soon: interactive prompt file filler.
    """
    sources = get_ordered_sources()

    # Determine scaffold_dict depending on index or template_choice
    if index is not None:
        if index < 1 or index > len(sources):
            raise typer.Exit(f"Invalid index {index}. Run 'mulch order' to see valid indices.")
        selected_source = sources[index - 1]

        if selected_source == "FALLBACK_SCAFFOLD":
            typer.echo(f"Using scaffold from embedded fallback (order index {index}).")
            scaffold_dict = FALLBACK_SCAFFOLD
        else:
            # Compose path to mulch.toml and load if exists
            scaffold_path = (target_dir / selected_source) if not Path(selected_source).is_absolute() else Path(selected_source)
            scaffold_path = (scaffold_path / "mulch.toml").resolve()

            if scaffold_path.exists():
                typer.echo(f"Using scaffold from file: {scaffold_path} (order index {index})")
                with open(scaffold_path, "r", encoding="utf-8") as f:
                    scaffold_dict = toml.load(f)
            else:
                raise typer.Exit(f"No scaffold file found at {scaffold_path}")

    elif template_choice:
        typer.secho(f"Choosing scaffold by template (choose from options)", fg=typer.colors.WHITE)
        template_choice_dict = load_template_choice_dictionary_from_file()
        scaffold_dict = template_choice_dict[template_choice]  # Make sure template_choice is a valid key
        path = "DEVELOPMENT: template_choice is not implemented yet, but it will be a number 1-9."
        typer.echo(f"Using template from file: {path}")

    else:
        # Neither index nor template choice: pick first available scaffold from sources
        scaffold_dict = None
        for idx, source in enumerate(sources, start=1):
            if source == "FALLBACK_SCAFFOLD":
                typer.echo(f"Using scaffold from embedded fallback (order index {idx}).")
                scaffold_dict = FALLBACK_SCAFFOLD
                break
            path = (target_dir / source) if not Path(source).is_absolute() else Path(source)
            path = (path / "mulch.toml").resolve()
            if path.exists():
                typer.echo(f"Using scaffold from file: {path} (order index {idx})")
                with open(path, "r", encoding="utf-8") as f:
                    scaffold_dict = toml.load(f)
                break

        if scaffold_dict is None:
            raise typer.Exit("No available scaffold found in any source.")

    # Write scaffold to target_dir/.mulch/mulch.toml
    output_path = target_dir / ".mulch" / DEFAULT_SCAFFOLD_FILENAME
    if output_path.exists() and not typer.confirm(f"⚠️ {output_path} already exists. Overwrite?"):
        raise typer.Abort()
    output_path.parent.mkdir(parents=True, exist_ok=True)
    with open(output_path, "w", encoding="utf-8") as f:
        toml.dump(scaffold_dict, f)

    typer.echo(f"✅ Wrote .mulch to: {output_path}")

    if edit or typer.confirm("📝 Would you like to open the scaffold file for editing now?", default=True):
        open_editor(output_path)

    typer.secho(
        "✏️  You can now manually edit the folder contents to customize your workspace layout and other mulch configuration.",
        fg=typer.colors.WHITE,
    )
    typer.echo(
        "⚙️  Changes to the scaffold file will directly affect the workspace layout and the generated workspace_manager.py when you run 'mulch src'."
    )

    # this does not do anything, but it could be useful to format it this way
    #build_dotmulch_standard_contents(target_dir=target_dir, scaffold_dict = scaffold_dict)

#@with_logging(use_portable=True)
@app.command()
def order(target_dir: Path = typer.Option(Path.cwd(), "--target-dir", "-t")):
    """
    Show the ordered list of mulch scaffold search paths, and also the FALLBACK_SCAFFOLD variable, and indicate which exist.
    """
    console = Console()
    sources = get_ordered_sources()

    typer.echo(f"Mulch Scaffold Order of Respect:")
    for idx, source in enumerate(sources, start=1):
        if source == "FALLBACK_SCAFFOLD":
            display_path = "FALLBACK_SCAFFOLD"
            exists = True  # Always exists because embedded
            typer.echo(f"  {idx}: FALLBACK_SCAFFOLD  ({'exists' if exists else 'does not exist'})")

        else:
            base_path = (target_dir / source) if not Path(source).is_absolute() else Path(source)
            resolved_path = (base_path / "mulch.toml").resolve()
            display_path = str(resolved_path)
            exists = resolved_path.exists()
            typer.echo(f"  {idx}: {display_path} ({'exists' if exists else 'does not exist'})")

        
    table = Table(title="Mulch Scaffold Order of Respect")
    table.add_column("Index", justify="right", style="cyan", no_wrap=True)
    table.add_column("Path", style="magenta")
    table.add_column("Exists?", justify="center", style="green")

    for idx, source in enumerate(sources, start=1):
        if source == "FALLBACK_SCAFFOLD":
            display_path = "FALLBACK_SCAFFOLD"
            exists = True  # Always exists because embedded
        else:
            base_path = (target_dir / source) if not Path(source).is_absolute() else Path(source)
            resolved_path = (base_path / "mulch.toml").resolve()
            display_path = str(resolved_path)
            exists = resolved_path.exists()

        table.add_row(str(idx), display_path, "✅" if exists else "❌")

    console.print(table)

def get_ordered_sources():
    return list(dedupe_paths(ORDER_OF_RESPECT)) + ["FALLBACK_SCAFFOLD"]

@app.command()
def show(index: int = typer.Argument(None, help="Index from 'mulch order' to display"),
         target_dir: Path = typer.Option(Path.cwd(), "--target-dir", "-t")):
    """
    Show the scaffold from the first available source in the order of respect,
    or from the specific index if provided.
    """
    sources = get_ordered_sources()

    if index is not None:
        if index < 1 or index > len(sources):
            raise typer.Exit(f"Invalid index {index}. Run 'mulch order' to see valid indices.")

        selected = sources[index - 1]

        if selected == "FALLBACK_SCAFFOLD":
            scaffold = FALLBACK_SCAFFOLD  # your embedded dict
            print("Loaded scaffold from embedded fallback structure.")
            print(toml.dumps(scaffold))
            return

        path = (target_dir / selected) if not Path(selected).is_absolute() else Path(selected)
        path = (path / "mulch.toml").resolve()

        if path.exists():
            typer.echo(f"Loaded scaffold from: {path}")
            with open(path, "r", encoding="utf-8") as f:
                scaffold = toml.load(f)
            print(toml.dumps(scaffold))
            return
        else:
            raise typer.Exit(f"No scaffold file found at: {path}")

    else:
        # No index given: check all sources in order until one loads
        for source in sources:
            if source == "FALLBACK_SCAFFOLD":
                scaffold = FALLBACK_SCAFFOLD
                print("Loaded scaffold from embedded fallback structure.")
                print(toml.dumps(scaffold))
                return

            path = (target_dir / source) if not Path(source).is_absolute() else Path(source)
            path = (path / "mulch.toml").resolve()

            if path.exists():
                typer.echo(f"Loaded scaffold from: {path}")
                with open(path, "r", encoding="utf-8") as f:
                    scaffold = toml.load(f)
                print(toml.dumps(scaffold))
                return

        raise typer.Exit("No scaffold found in any source.")

@app.command()
def validate():
    pass

@app.command()
def switch():
    """
    Switch the current workspace to a different one.
    This is a placeholder for future implementation.
    """
    typer.secho("This command is not yet implemented.", fg=typer.colors.YELLOW)
    typer.echo("You can switch workspaces by manually changing the workspace directory in your code editor or terminal.")
    typer.echo("Future versions of Mulch may support this functionality directly.")

if __name__ == "__main__":
    app()



from pathlib import Path
class PathContext:
    def __init__(self,
                 base_path: Path,
                 project_name: str = None,
                 workspace_name: str = None,
                 *,
                 here: bool = False,
                 stealth: bool = False):
        self.base_path = base_path
        self.project_name = project_name
        self.workspace_name = workspace_name
        self.here = here
        self.stealth = stealth

    @property
    def workspace_dir(self) -> Path:
        if self.here or self.stealth:
            return self.base_path / self.workspace_name
        return self.base_path / "workspaces" / self.workspace_name

    @property
    def source_dir(self) -> Path:
        if self.stealth:
            return self.base_path / ".mulch" / "src"
        return self.base_path / "src"

    @property
    def module_dir(self) -> Path:
        if self.project_name is None:
            self.project_name = self.base_path.name
        return self.source_dir / self.project_name

    @property
    def manager_path(self) -> Path:
        return self.module_dir / "workspace_manager.py"
    
    @property
    def manager_lock_path(self) -> Path:
        return self.module_dir / "manager.lock"
    
    @property
    def workspace_lock_path(self) -> Path:
        return self.workspace_dir / "space.lock"
    
    @property
    def flags_lock_path(self) -> Path:
        return self.base_path / ".mulch" / "flags.lock"


# src/mulch/workspace_instance_factory.py (project-agnostic, reusable)

import json
import logging
from pathlib import Path
from jinja2 import Environment, PackageLoader, select_autoescape #,FileSystemLoader
import toml
import typer
from importlib.resources import files
from typing import Dict, Any, Optional

from mulch.helpers import get_global_config_path, get_user_root, try_load_scaffold_file
from mulch.constants import FALLBACK_SCAFFOLD, DEFAULT_SCAFFOLD_FILENAME
from mulch.logging_setup import setup_logging, setup_logging_portable
from mulch.flags_log_status import FlagsLockStatus
from mulch.workspace_status import WorkspaceStatus
from mulch.basepath_manager import PathContext


setup_logging_portable()
logger = logging.getLogger(__name__)


class WorkspaceInstanceFactory:
    """
    Project-agnostic workspace instance factory for use with the mulch CLI.
    Manages directory creation and standardized file placement based on scaffold definition.
    """
    
    DEFAULT_WORKSPACE_CONFIG_FILENAME = "default-workspace.toml"
    DEFAULT_TEMPLATE_DIR = Path(__file__).parent / "templates"
    DEFAULT_WORKSPACE_TEMPLATE_FILENAME = "workspace_manager.py.j2"
    FALLBACK_SCAFFOLD = FALLBACK_SCAFFOLD # to make accessible, for pip and interally
    DEFAULT_SCAFFOLD_FILENAME = DEFAULT_SCAFFOLD_FILENAME # to make accessible, for pip and interally
    

    def __init__(self, 
                 base_path: Path,
                 workspaces_dir: Path,
                 workspace_name: str,
                 lock_data: dict,
                 *,
                 here: bool = False):
        """
        WorkspaceInstanceFactory.here is an attribute that references the setting of the --here flag, a flag of the command 'mulch workspace'. 
        When a workspace is generated, the value of self.here determines whether the workspace is generated in the root dir (if here is True, if the --here flag is used) or in root/workspaces/ (the default, if self.here is False, if the --here flag is not used.). 
        All workspaces should go in the same place for a given project. Ergo, to ensure this, a lock file should be set for the first workspace instance generated.
        
        Initialize a new workspace instance factory.
        
        Args:
            base_path: Root directory of the project
            workspace_dir: Directory where workspace will be created
            workspace_name: Name of the workspace
            lock_data: Scaffold and metadata for lockfile
            here: If True, create workspace in root instead of workspaces/
            stealth: If True, use .mulch/workspaces/ instead of root/workspaces/ # deprecated, blocked, no way to use.
        
        """
        self.base_path = Path(base_path).resolve()
        self.workspace_name = workspace_name
        self.lock_data = lock_data
        self.here = here
        self.stealth = False  # deprecated, blocked, no way to use. 
        
        # Create PathContext with explicit parameters
        self.context = PathContext(
            base_path=self.base_path,
            workspace_name=self.workspace_name,
            #workspace_dir=workspace_dir,
            here=self.here,
            stealth=self.stealth
        )

        # Get paths from context
        #self.workspace_dir = self.context.workspace_dir 
        self.workspace_dir = workspaces_dir / workspace_name # relies on WorkspaceInstanceFactory.determine_workspaces_dir() rather than PathContext. 
        self.workspace_lock_path = self.context.workspace_lock_path
        self.flags_lock_path = self.context.flags_lock_path


        self.project_name = self.base_path.name

    def get_path(self, key: str) -> Path:
        """
        Generic path getter using slash-separated key within the workspace.
        """
        path = self.workspace_dir
        for part in key.strip("/").split("/"):
            path /= part
        return path
    
    def evaluate_flags_lock_status(self) -> FlagsLockStatus:
        if not self.flags_lock_path.exists():
            return FlagsLockStatus.MISSING
        else:
            try:
                with open(self.flags_lock_path, "r", encoding="utf-8") as f:
                    existing = toml.loads(f)
            except:
                pass

    def evaluate_workspace_status(self) -> WorkspaceStatus:
        
        if not self.workspace_dir.exists():
            return WorkspaceStatus.MISSING

        if self.workspace_lock_path.exists():
            try:
                with open(self.workspace_lock_path, "r", encoding="utf-8") as f:
                    existing = json.load(f)
                existing_scaffold = existing.get("scaffold", {})
                if existing_scaffold == self.lock_data.get("scaffold", {}):
                    return WorkspaceStatus.MATCHES
                else:
                    return WorkspaceStatus.DIFFERS
            except Exception as e:
                logger.warning(f"Failed to read {self.workspace_lock_path}: {e}")
                return WorkspaceStatus.DIFFERS
        else:
            return WorkspaceStatus.EXISTS_NO_LOCK
        

        
    def write_workspace_lockfile(self) -> bool:
        
        with open(self.workspace_lock_path, "w", encoding="utf-8") as f:
            json.dump(self.lock_data, f, indent=2)
        logger.debug(f"Wrote lockfile to: {self.workspace_lock_path}")
    def write_workspace_lockfile(self):
        """Write workspace lock file after ensuring parent directory exists."""
        try:
            # Create workspace directory
            self.workspace_dir.mkdir(parents=True, exist_ok=True)
            
            # Write the lock file
            with open(self.workspace_lock_path, "w", encoding="utf-8") as f:
                json.dump(self.lock_data, f, indent=2)
            logger.debug(f"Wrote lockfile to: {self.workspace_lock_path}")
            return True
        except Exception as e:
            logger.error(f"Failed to write workspace lock file: {e}")
            return False

    @classmethod
    def determine_workspaces_dir(cls, target_dir: Path, here: bool) -> Path:
        """
        Determine the workspace directory path based on flags.
        
        Args:
            target_dir: Base project directory
            here: If True, create in current directory
            stealth: If True, use .mulch/workspaces/ # deprecated
        """
        if here:
            return target_dir 
        #if stealth:
        #    return target_dir / ".mulch" / "workspaces" 
        return target_dir / "workspaces" 
    
    def check_and_create_workspace_dirs_from_scaffold(self, workspace_dir: Path) -> None:
        """
        Create workspace directory structure from arbitrarily nested scaffold definition.
        
        Expected TOML structure:
        [scaffold]
        dirs = [
            "data",
            "data/raw",
            "data/processed/monthly",
            "queries/historical/archive",
        ]
        files = [
            "queries/historical/default-queries.toml",
            "data/processed/monthly/README.md"
        ]
        """
        scaffold_data = self.lock_data.get("scaffold", {})
        if "scaffold" in scaffold_data:  # Handle double-nested case from your TOML
            scaffold_data = scaffold_data["scaffold"]

        # Handle directory creation first
        if "dirs" in scaffold_data:
            for dir_path in scaffold_data["dirs"]:
                full_path = workspace_dir / dir_path
                if not full_path.exists():
                    full_path.mkdir(parents=True, exist_ok=True)
                    logger.debug(f"Created directory: {full_path}")
        
        # Handle file creation second (ensures parent dirs exist)
        if "files" in scaffold_data:
            for file_path in scaffold_data["files"]:
                full_path = workspace_dir / file_path
                if not full_path.exists():
                    full_path.parent.mkdir(parents=True, exist_ok=True)
                    full_path.touch()
                    logger.debug(f"Created file: {full_path}")

        # Handle alternative nested dict format if present
        if "structure" in scaffold_data:
        #if isinstance(scaffold_data.get("structure", {}), dict):
            def create_nested_structure(parent_path: Path, structure: dict) -> None:
                """Recursively create nested directory structure"""
                for name, content in structure.items():
                    current_path = parent_path / name
                    
                    if isinstance(content, dict):
                        # It's a directory with nested content
                        current_path.mkdir(parents=True, exist_ok=True)
                        logger.debug(f"Created directory: {current_path}")
                        create_nested_structure(current_path, content)
                    elif isinstance(content, list):
                        # It's a directory with list of files/dirs
                        current_path.mkdir(parents=True, exist_ok=True)
                        logger.debug(f"Created directory: {current_path}")
                        for item in content:
                            item_path = current_path / item
                            if "." in item:  # Basic file detection
                                item_path.parent.mkdir(parents=True, exist_ok=True)
                                item_path.touch()
                                logger.debug(f"Created file: {item_path}")
                            else:
                                item_path.mkdir(parents=True, exist_ok=True)
                                logger.debug(f"Created directory: {item_path}")
                    elif content is None or content == "":
                        # It's an empty directory
                        current_path.mkdir(parents=True, exist_ok=True)
                        logger.debug(f"Created empty directory: {current_path}")
                    else:
                        # Assume it's a file
                        current_path.parent.mkdir(parents=True, exist_ok=True)
                        current_path.touch()
                        logger.debug(f"Created file: {current_path}")
            
            create_nested_structure(workspace_dir, scaffold_data["structure"])

        logger.info(f"Completed workspace scaffold creation in {workspace_dir}")

    def create_workspace(self, *, set_default: bool = True) -> None:
        """
        Create a new workspace with scaffold structure and configuration.
        
        Args:
            set_default: Whether to set this as the default workspace (ignored if --here flag was used)
        """
        # Check workspace status first
        status = self.evaluate_workspace_status()
        if status != WorkspaceStatus.MISSING:
            logger.warning(f"Workspace already exists with status: {status}")
            return

        # Create the workspace structure
        self.write_workspace_lockfile()
        self.check_and_create_workspace_dirs_from_scaffold(self.workspace_dir)
        self.build_scaffolded_workspace_files()

        # Handle default workspace setting
        if set_default and not self.here:
            self.create_default_workspace_toml(self.workspace_dir.parent, self.workspace_name)
        
        typer.secho(f"✨ Workspace created at: {self.workspace_dir}", fg=typer.colors.BRIGHT_GREEN)

    
    @classmethod
    def create_default_workspace_toml(cls, workspaces_root: Path, workspace_name: str) -> None:
        """
        Write default-workspace.toml to the workspaces directory.
        
        Args:
            workspaces_root: Directory containing all workspaces (usually 'workspaces/')
            workspace_name: Name of the workspace to set as default
        """
        config_path = workspaces_root / cls.DEFAULT_WORKSPACE_CONFIG_FILENAME
        if not config_path.exists():
            config_path.write_text(f"[default-workspace]\nworkspace = \"{workspace_name}\"\n")
            logger.debug(f"Created {config_path}")
        else:
            logging.debug(f"{config_path} already exists; skipping overwrite")

    def build_scaffolded_workspace_files(self) -> None:
        """
        Seed both static and templated workspace files.
        Call this after workspace creation.
        Seed only placeholder files that are already declared in scaffold and still empty.
        This ensures the scaffold drives structure, not the seeder.
        """
        self.build_static_workspace_files()
        self.build_templated_workspace_files()
        
    def build_static_workspace_files(self) -> None:
        """
        Populate essential workspace files *only if* their placeholder files already exist.
        Avoids introducing files/folders not declared in the scaffold.
        """
        seed_map = {
            #Path("secrets") / "secrets-example.yaml": "secrets-example.yaml",
            Path("queries") / "default-queries.toml": "default-queries.toml",
        }

        for rel_path, src_filename in seed_map.items():
            dest = self.workspace_dir / rel_path
            # Clarify that seeders depend on placeholders
            if dest.exists() and dest.stat().st_size == 0:
                try:
                    src = files("mulch") / src_filename
                    with src.open("r", encoding="utf-8") as f_in:
                        contents = f_in.read()
                    dest.write_text(contents, encoding="utf-8")
                    logger.debug(f"Seeded workspace file: {dest}")
                    typer.echo(f"Seeded workspace file: {dest.name}")
                except Exception as e:
                    logger.warning(f"Failed to seed {rel_path}: {e}")
            else:
                logger.debug(f"Skipped seeding {dest}; file doesn't exist or is not empty.")

    def build_templated_workspace_files(self):
        """
        Generate helpful default files in the new workspace, such as about_this_workspace.md.
        """
        workspace_dir = self.workspace_dir

        env = Environment(
            loader=PackageLoader("mulch", "templates"),
            autoescape=select_autoescape()
        )

        about_path = workspace_dir / "about_this_workspace.md"

        if not about_path.exists():
            try:
                template = env.get_template("about_this_workspace.md.j2")
                content = template.render(
                    workspace_name=self.workspace_name,
                    generated_at=self.lock_data.get("generated_at", ""),
                    scaffold_source=self.lock_data.get("generated_by", "")
                )
                about_path.write_text(content, encoding="utf-8")
            except Exception as e:
                logger.warning(f"Failed to render about_this_workspace.md from template: {e}")
                content = f"# About {self.workspace_name}\n\nGenerated on {self.lock_data.get('generated_at', '')}"
            logging.debug(f"Seeded {about_path}")
        else:
            logging.debug(f"{about_path} already exists; skipping")


def load_scaffold_(scaffold_path: Path | None = None) -> dict:
    if not scaffold_path:
        scaffold_path = Path(__file__).parent / DEFAULT_SCAFFOLD_FILENAME
    
    if not scaffold_path.exists():
        # File missing, log warning and return fallback
        typer.echo(f"Missing scaffold file, using fallback scaffold.")
        logger.debug(f"Warning: Missing scaffold file: {scaffold_path}, using fallback scaffold.")
        return FALLBACK_SCAFFOLD
        
    #with open(scaffold_path, "r") as f:
    #    return json.load(f)
        
    try:
        with open(scaffold_path, "r") as f:
            content = f.read().strip()
            if not content:
                logger.debug(f"Warning: Scaffold file {scaffold_path} is empty, using fallback scaffold.")
                typer.echo(f"Scaffold file is empty, using fallback scaffold.")
                return FALLBACK_SCAFFOLD
            return json.loads(content)
    except json.JSONDecodeError as e:
        logger.warning(f"Warning: Scaffold file {scaffold_path} contains invalid JSON ({e}), using fallback scaffold.")
        return FALLBACK_SCAFFOLD

def load_scaffold__(target_dir: Optional[Path] = None, 
                  strict_local_dotmulch:bool=False, 
                  seed_if_missing:bool=False) -> Dict[str, Any]:
    target_dir = target_dir or Path.cwd()
    base = target_dir / ".mulch"

    if strict_local_dotmulch:
        # Only try .mulch — no fallback
        for fname in filenames:
            path = base / fname
            scaffold = try_load_scaffold_file(path)
            if scaffold:
                logger.info(f"✅ Loaded scaffold from: {path}")
                return scaffold

        if seed_if_missing:
            from mulch.seed_logic import write_seed_scaffold  # or wherever this lives
            logger.warning("⚠️ .mulch exists but no scaffold file found. Auto-seeding...")
            write_seed_scaffold(target_dir)
            return load_scaffold(target_dir, strict_local_dotmulch=True, seed_if_missing=False)
        

        raise FileNotFoundError("🚫 No valid `.mulch/mulch-scaffold.*` found and auto-seed not enabled.")

    # Default behavior: search all fallback paths
    
    base_dirs = [
        target_dir / ".mulch",    # 1. Local .mulch folder
        target_dir,               # 2. Root project dir
        Path.home() / 'mulch',               # 3. User root on system
        get_global_config_path(appname = "mulch") # 4. Global config
    ]
    
    filenames = ["mulch.toml"]

    for base in base_dirs:
        for filename in filenames:
            path = base / filename
            scaffold = try_load_scaffold_file(path)
            if scaffold:
                logger.info(f"✅ Loaded scaffold from: {path}")
                return scaffold
            
    logger.warning("No valid scaffold file found. Falling back to internal scaffold.")
    return FALLBACK_SCAFFOLD

# Move outside class
def load_scaffold(
    target_dir: Optional[Path] = None, 
    strict_local_dotmulch: bool = False,
    seed_if_missing: bool = False
) -> Dict[str, Any]:
    """
    Load scaffold configuration from various possible locations.
    
    Args:
        target_dir: Directory to start searching from (defaults to cwd)
        strict_local_dotmulch: If True, only look in .mulch directory
        seed_if_missing: If True and strict mode, create seed scaffold
    
    Returns:
        Dict containing scaffold configuration
    
    Raises:
        FileNotFoundError: If strict mode and no scaffold found
    """
    target_dir = target_dir or Path.cwd()
    base = target_dir / ".mulch"

    filenames = ["mulch.toml"]

    if strict_local_dotmulch:
        # ...existing strict mode code...
        pass

    # Default behavior: search all fallback paths
    base_dirs = [
        target_dir / ".mulch",    # 1. Local .mulch folder
        target_dir,               # 2. Root project dir
        Path.home() / 'mulch',    # 3. User root on system
        get_global_config_path(appname="mulch")  # 4. Global config
    ]

    for base in base_dirs:
        for filename in filenames:
            path = base / filename
            scaffold = try_load_scaffold_file(path)
            if scaffold:
                logger.info(f"📄 Loaded scaffold from: {path}")
                return scaffold
            
    logger.warning("No valid scaffold file found. Using fallback scaffold.")
    return FALLBACK_SCAFFOLD


import toml
DEFAULT_SCAFFOLD_FILENAME = "mulch.toml"

LOCK_FILE_NAME = 'mulch.lock'

FALLBACK_SCAFFOLD_JSON = {
        "": ["scripts", "tools", "templates", "exports", "imports", "images", "documents", "configurations", "about_this_workspace.md"],
        "exports": ["aggregate"],
        "configurations": ["default-workspace.toml", "logging.json"],
        "secrets": ["to-be-run-at-creation.yaml"],
        "tools": ["to-be-used-as-needed.toml"]
    }

FALLBACK_SCAFFOLD_TOML = '''[scaffold]
dirs = [
    "data",
    "secrets",
    "data/raw",
    "data/processed/monthly",
    "queries/historical/archive",
]
files = [
    "queries/historical/default-queries.toml",
    "data/processed/monthly/README.md"
]'''
FALLBACK_SCAFFOLD = toml.loads(FALLBACK_SCAFFOLD_TOML)

'''EXAMPLE
data = {
    "section": {
        "key": "value",
        "list": [1, 2, 3],
        "subsection": {
            "foo": "bar"
        }
    }
} 

with open("file.toml", "w") as f:
    toml.dump(data, f)

[section]
key = "value"
list = [1, 2, 3]

[section.subsection]
foo = "bar"
'''


# mulch/reg-backup.py
import os

# Use LocalAppData for script location
local_app_data = os.environ['LOCALAPPDATA']
call_script = os.path.join(local_app_data, 'mulch', 'call-mulch-workspace.ps1')

# Commands for context menus
background_command = f'powershell.exe -NoProfile -ExecutionPolicy Bypass -File "{call_script}" "%V"'
folder_command = f'powershell.exe -NoProfile -ExecutionPolicy Bypass -File "{call_script}" "%L"'

# Backup .reg content
reg_content = f"""Windows Registry Editor Version 5.00

; Right-click background of a folder
[HKEY_CURRENT_USER\\Software\\Classes\\Directory\\Background\\shell\\mulch_workspace]
@="mulch workspace"
"Icon"="%LOCALAPPDATA%\\\\mulch\\\\mulch-icon.ico"
"Position"="Top"
"CommandFlags"=dword:00000000
[HKEY_CURRENT_USER\\Software\\Classes\\Directory\\Background\\shell\\mulch_workspace\\command]
@="{background_command}"

; Right-click ON folder
[HKEY_CURRENT_USER\\Software\\Classes\\Directory\\shell\\mulch_workspace]
@="mulch workspace"
"Icon"="%LOCALAPPDATA%\\\\mulch\\\\mulch-icon.ico"
"Position"="Top"
"CommandFlags"=dword:00000000
[HKEY_CURRENT_USER\\Software\\Classes\\Directory\\shell\\mulch_workspace\\command]
@="{folder_command}"
"""

# Ensure backup folder exists
backup_dir = os.path.join(local_app_data, 'mulch')
os.makedirs(backup_dir, exist_ok=True)

# Write the .reg file
backup_path = os.path.join(backup_dir, "mulch_context_backup.reg")
with open(backup_path, "w", encoding="utf-16") as f:
    f.write(reg_content)

print(f"Backup .reg file created at: {backup_path}")



DOTMULCH_STANDARD_SCAFFOLD = {
        "": ["scripts", "tools", "templates", "configurations", "about_this_workspace.md"],
        "configurations": ["default-workspace.toml", "logging.json"],
        "scripts": ["to-be-run-at-creation.yaml"],
        "tools": ["to-be-used-as-needed.toml"]
    }
        
def build_dotmulch_standard_contents(target_dir, scaffold_dict = None):
    if scaffold_dict is None:
        scaffold_dict = DOTMULCH_STANDARD_SCAFFOLD
        


import json
from pathlib import Path
from datetime import datetime

REFERENCE_LOCK_PATH = Path(".mulch/reference.lock")

def load_reference_lock() -> dict:
    if REFERENCE_LOCK_PATH.exists():
        with open(REFERENCE_LOCK_PATH, "r", encoding="utf-8") as f:
            return json.load(f)
    return {
        "workspaces": {"instances": []},
        "src": {"instances": []},
        "validation": {"is_consistent": True, "issues": []},
        "metadata": {"workspace_updated": None, "src_updated": None, "version": "0.1"},
    }

def build_flags_record(here: bool = False, stealth: bool = False, force: bool = False, 
                       name: str = None, pattern: str = None,expliref: str = None) -> list[str]:
    flags = []
    if here:
        flags.append("--here")
    if stealth:
        flags.append("--stealth")
    if force:
        flags.append("--force")
    if name:
        flags.append("--name")
    if pattern:
        flags.append("--pattern")
    if expliref:
        flags.append("--expliref")
    return flags


REFERENCE_LOCK_PATH = Path(".mulch/reference.lock")

    
def save_reference_lock(data: dict):
    REFERENCE_LOCK_PATH.parent.mkdir(parents=True, exist_ok=True)
    with open(REFERENCE_LOCK_PATH, "w", encoding="utf-8") as f:
        json.dump(data, f, indent=2)

def validate_reference_lock(data: dict) -> dict:
    issues = []
    ws_paths = data.get("workspaces", {}).get("instances", [])
    src_paths = data.get("src", {}).get("instances", [])

    if not ws_paths:
        issues.append("No workspaces registered in reference.lock.")
    if not src_paths:
        issues.append("No src instances registered in reference.lock.")

    # Additional validation...

    data["validation"]["is_consistent"] = len(issues) == 0
    data["validation"]["issues"] = issues
    return data


class ReferenceLockManager:

    @staticmethod
    def load_lock() -> dict:
        return load_reference_lock()

    @staticmethod
    def save_lock(data: dict):
        save_reference_lock(data)

    @staticmethod
    def update_lock_workspace(pathstr: str, command_line: str, flags: list[str]) -> dict:
        """ Update or add a workspace entry in the reference lock file. """

        # use pathstr instead of path to avoid Path serialization issues, and to allow "null" as a string
        path = str(pathstr) # in case a Path object is passed, but can also be "null"
        data = load_reference_lock()
        now_iso = datetime.utcnow().isoformat() + "Z"

        workspaces = data.setdefault("workspaces", {})
        instances = workspaces.setdefault("instances", [])

        # Update existing or append new workspace entry by path
        for ws in instances:
            if ws["path"] == path:
                ws["flags"] = flags
                ws["command_line"] = command_line
                break
        else:
            instances.append({"path": path, 
                              "flags": flags,
                              "command_line": command_line})

        data.setdefault("metadata", {})
        data["metadata"]["workspace_updated"] = now_iso

        data = validate_reference_lock(data)
        save_reference_lock(data)
        return data

    @staticmethod
    def update_lock_src(pathstr: str, command_line: str, flags: list[str]) -> dict:
        path = str(pathstr)
        data = load_reference_lock()
        now_iso = datetime.utcnow().isoformat() + "Z"

        src = data.setdefault("src", {})
        instances = src.setdefault("instances", [])

        # Update existing or append new src entry by path
        for sourcepath in instances:
            if sourcepath["path"] == path:
                sourcepath["flags"] = flags
                sourcepath["command_line"] = command_line
                break
        else:
            instances.append({"path": path,
                              "flags": flags,
                              "command_line": command_line})

        data.setdefault("metadata", {})
        data["metadata"]["src_updated"] = now_iso

        data = validate_reference_lock(data)
        save_reference_lock(data)
        return data
    
    @staticmethod
    def list_workspaces() -> list[dict]:
        data = load_reference_lock()
        return data.get("workspaces", {}).get("instances", [])

    @staticmethod
    def list_src() -> list[dict]:
        data = load_reference_lock()
        return data.get("src", {}).get("instances", [])


# src/mulch/active_workspace_manager.py

import toml
from pathlib import Path
from datetime import datetime

ACTIVE_WORKSPACE_PATH = Path(".mulch/active_workspace.toml")

def load_active_workspace() -> dict:
    if ACTIVE_WORKSPACE_PATH.exists():
        return toml.load(ACTIVE_WORKSPACE_PATH)
    return {"path": None, "activated_at": None}

def save_active_workspace(path: str):
    ACTIVE_WORKSPACE_PATH.parent.mkdir(parents=True, exist_ok=True)
    data = {
        "path": path,
        "activated_at": datetime.utcnow().isoformat() + "Z",
    }
    with open(ACTIVE_WORKSPACE_PATH, "w", encoding="utf-8") as f:
        toml.dump(data, f)

def clear_active_workspace():
    if ACTIVE_WORKSPACE_PATH.exists():
        ACTIVE_WORKSPACE_PATH.unlink()



# src/mulch/scaffold_loader.py

from pathlib import Path
from typing import Dict, Any, Optional, List
import toml
import json
import logging

from mulch.constants import FALLBACK_SCAFFOLD, DEFAULT_SCAFFOLD_FILENAME
from mulch.commands.dotfolder import create_dot_mulch
logger = logging.getLogger(__name__)

def load_scaffold_file(path: Path) -> Optional[Dict[str, Any]]:
    """Load scaffold configuration from file."""
    if not path.exists():
        return None
        
    try:
        return toml.load(path)
    except Exception as e:
        logger.warning(f"Failed to parse {path}: {e}")
        return None

def resolve_scaffold(
    search_paths: List[Path],
    filenames: List[str]
) -> Dict[str, Any]:
    """
    High-level function to resolve scaffold across multiple possible locations.
    Follows order of precedence rules.
    """
    for base in search_paths:
        for filename in filenames:
            path = base / filename
            data = load_scaffold_file(path)
            if data:
                logger.info(f"📄 Loaded scaffold from: {path}")
                return data
    
    logger.warning("No valid scaffold file found, using fallback")
    return FALLBACK_SCAFFOLD  # Now uses the TOML-based scaffold
